package dht

import (
	"context"
	"time"

	"gitlab.dms3.io/p2p/go-p2p-core/peer"

	"gitlab.dms3.io/mf/go-multiaddr"
)

// DefaultBootstrapPeers is a set of public DHT bootstrap peers provided by p2p.
var DefaultBootstrapPeers []multiaddr.Multiaddr

// Minimum number of peers in the routing table. If we drop below this and we
// see a new peer, we trigger a bootstrap round.
var minRTRefreshThreshold = 10

const (
	periodicBootstrapInterval = 2 * time.Minute
	maxNBoostrappers          = 2
)

func init() {
	for _, s := range []string{
		"/dnsaddr/bootstrap.p2p.io/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m", // FIXME TODO: replace with dms peer.IDs
		"/dnsaddr/bootstrap.p2p.io/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m",
		"/dnsaddr/bootstrap.p2p.io/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m",
		"/dnsaddr/bootstrap.p2p.io/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m",
		"/ip4/104.131.131.82/tcp/4001/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m", // mars.i.ip fs.io
	} {
		ma, err := multiaddr.NewMultiaddr(s)
		if err != nil {
			panic(err)
		}
		DefaultBootstrapPeers = append(DefaultBootstrapPeers, ma)
	}
}

// GetDefaultBootstrapPeerAddrInfos returns the peer.AddrInfos for the default
// bootstrap peers so we can use these for initializing the DHT by passing these to the
// BootstrapPeers(...) option.
func GetDefaultBootstrapPeerAddrInfos() []peer.AddrInfo {
	ds := make([]peer.AddrInfo, 0, len(DefaultBootstrapPeers))

	for i := range DefaultBootstrapPeers {
		info, err := peer.AddrInfoFromP2pAddr(DefaultBootstrapPeers[i])
		if err != nil {
			logger.Errorw("failed to convert bootstrapper address to peer addr info", "address",
				DefaultBootstrapPeers[i].String(), err, "err")
			continue
		}
		ds = append(ds, *info)
	}
	return ds
}

// Bootstrap tells the DHT to get into a bootstrapped state satisfying the
// Dms3Router interface.
func (dht *Dms3DHT) Bootstrap(ctx context.Context) error {
	dht.fixRTIfNeeded()
	dht.rtRefreshManager.RefreshNoWait()
	return nil
}

// RefreshRoutingTable tells the DHT to refresh it's routing tables.
//
// The returned channel will block until the refresh finishes, then yield the
// error and close. The channel is buffered and safe to ignore.
func (dht *Dms3DHT) RefreshRoutingTable() <-chan error {
	return dht.rtRefreshManager.Refresh(false)
}

// ForceRefresh acts like RefreshRoutingTable but forces the DHT to refresh all
// buckets in the Routing Table irrespective of when they were last refreshed.
//
// The returned channel will block until the refresh finishes, then yield the
// error and close. The channel is buffered and safe to ignore.
func (dht *Dms3DHT) ForceRefresh() <-chan error {
	return dht.rtRefreshManager.Refresh(true)
}
